# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import sushy

from ironic.common import boot_devices
from ironic.common import exception
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conductor import utils as manager_utils
from ironic.drivers.modules import boot_mode_utils
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules import image_utils
from ironic.drivers.modules.redfish import boot as redfish_boot
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils
from ironic.tests.unit.objects import utils as obj_utils

INFO_DICT = db_utils.get_test_redfish_info()


@mock.patch('oslo_utils.eventletutils.EventletEvent.wait',
            lambda *args, **kwargs: None)
class RedfishVirtualMediaBootTestCase(db_base.DbTestCase):

    def setUp(self):
        super(RedfishVirtualMediaBootTestCase, self).setUp()
        self.config(enabled_hardware_types=['redfish'],
                    enabled_power_interfaces=['redfish'],
                    enabled_boot_interfaces=['redfish-virtual-media'],
                    enabled_management_interfaces=['redfish'],
                    enabled_inspect_interfaces=['redfish'],
                    enabled_bios_interfaces=['redfish'])
        self.node = obj_utils.create_test_node(
            self.context, driver='redfish', driver_info=INFO_DICT)

    @mock.patch.object(deploy_utils, 'get_boot_option', lambda node: 'ramdisk')
    def test_parse_driver_info_ramdisk(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.driver_info = {}
            task.node.automated_clean = False
            actual_driver_info = redfish_boot._parse_driver_info(task.node)
            self.assertEqual({'can_provide_config': False},
                             actual_driver_info)

    def test_parse_driver_info_deploy(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            actual_driver_info = redfish_boot._parse_driver_info(task.node)

            self.assertIn('kernel', actual_driver_info['deploy_kernel'])
            self.assertIn('ramdisk', actual_driver_info['deploy_ramdisk'])
            self.assertIn('bootloader', actual_driver_info['bootloader'])
            self.assertTrue(actual_driver_info['can_provide_config'])

    def test_parse_driver_info_iso(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.driver_info.update(
                {'deploy_iso': 'http://boot.iso'})

            actual_driver_info = redfish_boot._parse_driver_info(task.node)

            self.assertEqual('http://boot.iso',
                             actual_driver_info['deploy_iso'])
            self.assertFalse(actual_driver_info['can_provide_config'])

    def test_parse_driver_info_iso_deprecated(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.driver_info.update(
                {'redfish_deploy_iso': 'http://boot.iso'})

            actual_driver_info = redfish_boot._parse_driver_info(task.node)

            self.assertEqual('http://boot.iso',
                             actual_driver_info['deploy_iso'])
            self.assertFalse(actual_driver_info['can_provide_config'])

    def test_parse_driver_info_removable(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.driver_info.update(
                {'deploy_iso': 'http://boot.iso',
                 'config_via_removable': True}
            )

            actual_driver_info = redfish_boot._parse_driver_info(task.node)
            self.assertTrue(actual_driver_info['config_via_removable'])
            self.assertTrue(actual_driver_info['can_provide_config'])

    @mock.patch.object(redfish_boot.LOG, 'warning', autospec=True)
    def test_parse_driver_info_removable_deprecated(self, mock_log):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader',
                 'config_via_floppy': True}
            )

            actual_driver_info = redfish_boot._parse_driver_info(task.node)
            self.assertTrue(actual_driver_info['config_via_removable'])
            self.assertTrue(mock_log.called)

    def test_parse_driver_info_rescue(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.RESCUING
            task.node.driver_info.update(
                {'rescue_kernel': 'kernel',
                 'rescue_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            actual_driver_info = redfish_boot._parse_driver_info(task.node)

            self.assertIn('kernel', actual_driver_info['rescue_kernel'])
            self.assertIn('ramdisk', actual_driver_info['rescue_ramdisk'])
            self.assertIn('bootloader', actual_driver_info['bootloader'])

    def test_parse_driver_info_exc(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.MissingParameterValue,
                              redfish_boot._parse_driver_info,
                              task.node)

    def _test_parse_driver_info_from_conf(self, mode='deploy', by_arch=False):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            if mode == 'rescue':
                task.node.provision_state = states.RESCUING

            if by_arch:
                ramdisk = 'glance://%s_ramdisk_uuid' % mode
                kernel = 'glance://%s_kernel_uuid' % mode

                config = {
                    '%s_ramdisk_by_arch' % mode: {'x86_64': ramdisk},
                    '%s_kernel_by_arch' % mode: {'x86_64': kernel}
                }
                expected = {
                    '%s_ramdisk' % mode: ramdisk,
                    '%s_kernel' % mode: kernel
                }
            else:
                expected = {
                    '%s_ramdisk' % mode: 'glance://%s_ramdisk_uuid' % mode,
                    '%s_kernel' % mode: 'glance://%s_kernel_uuid' % mode
                }
                config = expected

            self.config(group='conductor', **config)

            image_info = redfish_boot._parse_driver_info(task.node)

            for key, value in expected.items():
                self.assertEqual(value, image_info[key])

    def test_parse_driver_info_from_conf_deploy(self):
        self._test_parse_driver_info_from_conf()

    def test_parse_driver_info_from_conf_rescue(self):
        self._test_parse_driver_info_from_conf(mode='rescue')

    def test_parse_driver_info_from_conf_deploy_by_arch(self):
        self._test_parse_driver_info_from_conf(by_arch=True)

    def test_parse_driver_info_from_conf_rescue_by_arch(self):
        self._test_parse_driver_info_from_conf(mode='rescue', by_arch=True)

    def _test_parse_driver_info_mixed_source(self, mode='deploy',
                                             by_arch=False):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            if mode == 'rescue':
                task.node.provision_state = states.RESCUING

            if by_arch:
                kernel_config = {
                    '%s_kernel_by_arch' % mode: {
                        'x86': 'glance://%s_kernel_uuid' % mode
                    }
                }
            else:
                kernel_config = {
                    '%s_kernel' % mode: 'glance://%s_kernel_uuid' % mode
                }

            ramdisk_config = {
                '%s_ramdisk' % mode: 'glance://%s_ramdisk_uuid' % mode,
            }

            self.config(group='conductor', **kernel_config)

            task.node.driver_info.update(ramdisk_config)

            self.assertRaises(exception.MissingParameterValue,
                              redfish_boot._parse_driver_info, task.node)

    def test_parse_driver_info_mixed_source_deploy(self):
        self._test_parse_driver_info_mixed_source()

    def test_parse_driver_info_mixed_source_rescue(self):
        self._test_parse_driver_info_mixed_source(mode='rescue')

    def test_parse_driver_info_mixed_source_deploy_by_arch(self):
        self._test_parse_driver_info_mixed_source(by_arch=True)

    def test_parse_driver_info_mixed_source_rescue_by_arch(self):
        self._test_parse_driver_info_mixed_source(mode='rescue', by_arch=True)

    def _test_parse_driver_info_choose_by_arch(self, mode='deploy'):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            if mode == 'rescue':
                task.node.provision_state = states.RESCUING
            task.node.properties['cpu_arch'] = 'aarch64'
            wrong_ramdisk = 'glance://wrong_%s_ramdisk_uuid' % mode
            wrong_kernel = 'glance://wrong_%s_kernel_uuid' % mode
            ramdisk = 'glance://%s_ramdisk_uuid' % mode
            kernel = 'glance://%s_kernel_uuid' % mode

            config = {
                '%s_ramdisk_by_arch' % mode: {
                    'x86_64': wrong_ramdisk, 'aarch64': ramdisk},
                '%s_kernel_by_arch' % mode: {
                    'x86_64': wrong_kernel, 'aarch64': kernel}
            }
            expected = {
                '%s_ramdisk' % mode: ramdisk,
                '%s_kernel' % mode: kernel
            }

            self.config(group='conductor', **config)

            image_info = redfish_boot._parse_driver_info(task.node)

            for key, value in expected.items():
                self.assertEqual(value, image_info[key])

    def test_parse_driver_info_choose_by_arch_deploy(self):
        self._test_parse_driver_info_choose_by_arch()

    def test_parse_driver_info_choose_by_arch_rescue(self):
        self._test_parse_driver_info_choose_by_arch(mode='rescue')

    def _test_parse_driver_info_choose_by_hierarchy(self, mode='deploy',
                                                    ramdisk_missing=False):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            if mode == 'rescue':
                task.node.provision_state = states.RESCUING

            ramdisk = 'glance://def_%s_ramdisk_uuid' % mode
            kernel = 'glance://def_%s_kernel_uuid' % mode
            ramdisk_by_arch = 'glance://%s_ramdisk_by_arch_uuid' % mode
            kernel_by_arch = 'glance://%s_kernel_by_arch_uuid' % mode

            config = {
                '%s_kernel_by_arch' % mode: {
                    'x86_64': kernel_by_arch},
                '%s_ramdisk' % mode: ramdisk,
                '%s_kernel' % mode: kernel
            }
            if not ramdisk_missing:
                config['%s_ramdisk_by_arch' % mode] = {
                    'x86_64': ramdisk_by_arch}
                expected = {
                    '%s_ramdisk' % mode: ramdisk_by_arch,
                    '%s_kernel' % mode: kernel_by_arch
                }
            else:
                expected = {
                    '%s_ramdisk' % mode: ramdisk,
                    '%s_kernel' % mode: kernel
                }

            self.config(group='conductor', **config)

            image_info = redfish_boot._parse_driver_info(task.node)

            for key, value in expected.items():
                self.assertEqual(value, image_info[key])

    def test_parse_driver_info_choose_by_hierarchy_deploy(self):
        self._test_parse_driver_info_choose_by_hierarchy()

    def test_parse_driver_info_choose_by_hierarchy_rescue(self):
        self._test_parse_driver_info_choose_by_hierarchy(mode='rescue')

    def test_parse_driver_info_choose_by_hierarchy_missing_param_deploy(self):
        self._test_parse_driver_info_choose_by_hierarchy(ramdisk_missing=True)

    def test_parse_driver_info_choose_by_hierarchy_missing_param_rescue(self):
        self._test_parse_driver_info_choose_by_hierarchy(
            mode='rescue', ramdisk_missing=True)

    def test_parse_deploy_info(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.node.instance_info.update(
                {'image_source': 'http://boot/iso',
                 'kernel': 'http://kernel/img',
                 'ramdisk': 'http://ramdisk/img'})

            actual_instance_info = redfish_boot._parse_deploy_info(task.node)

            self.assertEqual(
                'http://boot/iso', actual_instance_info['image_source'])
            self.assertEqual(
                'http://kernel/img', actual_instance_info['kernel'])
            self.assertEqual(
                'http://ramdisk/img', actual_instance_info['ramdisk'])

    def test_parse_deploy_info_exc(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.MissingParameterValue,
                              redfish_boot._parse_deploy_info,
                              task.node)

    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    def test_validate_local(self, mock_parse_driver_info):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.instance_info = {}

            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.driver.boot.validate(task)

    @mock.patch.object(deploy_utils, 'get_boot_option', lambda node: 'ramdisk')
    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    @mock.patch.object(deploy_utils, 'validate_image_properties',
                       autospec=True)
    def test_validate_kernel_ramdisk(self, mock_validate_image_properties,
                                     mock_parse_driver_info):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.instance_info.update(
                {'kernel': 'kernel',
                 'ramdisk': 'ramdisk',
                 'image_source': 'http://image/source'}
            )

            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.driver.boot.validate(task)

            mock_validate_image_properties.assert_called_once_with(
                task, mock.ANY)

    @mock.patch.object(deploy_utils, 'get_boot_option', lambda node: 'ramdisk')
    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    @mock.patch.object(deploy_utils, 'validate_image_properties',
                       autospec=True)
    def test_validate_boot_iso(self, mock_validate_image_properties,
                               mock_parse_driver_info):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.instance_info.update(
                {'boot_iso': 'http://localhost/file.iso'}
            )

            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.driver.boot.validate(task)

            mock_validate_image_properties.assert_called_once_with(
                task, mock.ANY)

    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    @mock.patch.object(deploy_utils, 'validate_image_properties',
                       autospec=True)
    def test_validate_correct_vendor(self, mock_validate_image_properties,
                                     mock_parse_driver_info):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.instance_info.update(
                {'kernel': 'kernel',
                 'ramdisk': 'ramdisk',
                 'image_source': 'http://image/source'}
            )

            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.node.properties['vendor'] = "Ironic Co."

            task.driver.boot.validate(task)

    def test__validate_vendor_incompatible_with_idrac(self):
        managers = [mock.Mock(firmware_version='5.10.30.00',
                              manager_type=sushy.MANAGER_TYPE_BMC)]
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.instance_info.update(
                {'kernel': 'kernel',
                 'ramdisk': 'ramdisk',
                 'image_source': 'http://image/source'}
            )

            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.node.properties['vendor'] = "Dell Inc."

            self.assertRaisesRegex(
                exception.InvalidParameterValue, "with vendor Dell Inc.",
                task.driver.boot._validate_vendor, task, managers)

    def test__validate_vendor_compatible_with_idrac(self):
        managers = [mock.Mock(firmware_version='6.00.00.00',
                              manager_type=sushy.MANAGER_TYPE_BMC)]
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.instance_info.update(
                {'kernel': 'kernel',
                 'ramdisk': 'ramdisk',
                 'image_source': 'http://image/source'}
            )

            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.node.properties['vendor'] = "Dell Inc."

            task.driver.boot._validate_vendor(task, managers)

    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    @mock.patch.object(deploy_utils, 'validate_image_properties',
                       autospec=True)
    def test_validate_missing(self, mock_validate_image_properties,
                              mock_parse_driver_info):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.MissingParameterValue,
                              task.driver.boot.validate, task)

    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    def test_validate_inspection(self, mock_parse_driver_info):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.driver.boot.validate_inspection(task)

            mock_parse_driver_info.assert_called_once_with(task.node)

    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    def test_validate_inspection_missing(self, mock_parse_driver_info):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.MissingParameterValue,
                              task.driver.boot.validate_inspection, task)

    @mock.patch.object(redfish_boot.RedfishVirtualMediaBoot,
                       '_validate_vendor', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(image_utils, 'prepare_deploy_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_insert_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_driver_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_power_action',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_ramdisk_with_params(
            self, mock_system, mock_boot_mode_utils, mock_node_power_action,
            mock__parse_driver_info, mock__insert_vmedia, mock__eject_vmedia,
            mock_prepare_deploy_iso, mock_node_set_boot_device,
            mock_validate_vendor):

        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.node.provision_state = states.DEPLOYING

            mock__parse_driver_info.return_value = {}
            mock_prepare_deploy_iso.return_value = 'image-url'

            task.driver.boot.prepare_ramdisk(task, {})

            mock_validate_vendor.assert_called_once_with(
                task.driver.boot, task, managers
            )

            mock_node_power_action.assert_called_once_with(
                task, states.POWER_OFF)

            mock__eject_vmedia.assert_called_once_with(
                task, managers, sushy.VIRTUAL_MEDIA_CD)

            mock__insert_vmedia.assert_called_once_with(
                task, managers, 'image-url', sushy.VIRTUAL_MEDIA_CD)

            token = task.node.driver_internal_info['agent_secret_token']
            self.assertTrue(token)

            expected_params = {
                'ipa-agent-token': token,
                'ipa-debug': '1',
                'boot_method': 'vmedia',
            }

            mock_prepare_deploy_iso.assert_called_once_with(
                task, expected_params, 'deploy', {})

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.CDROM, False)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)

            self.assertTrue(task.node.driver_internal_info[
                'agent_secret_token_pregenerated'])

    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(image_utils, 'prepare_deploy_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_insert_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_driver_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_power_action',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_ramdisk_no_debug(
            self, mock_system, mock_boot_mode_utils, mock_node_power_action,
            mock__parse_driver_info, mock__insert_vmedia, mock__eject_vmedia,
            mock_prepare_deploy_iso, mock_node_set_boot_device):
        self.config(debug=False)
        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING

            mock__parse_driver_info.return_value = {}
            mock_prepare_deploy_iso.return_value = 'image-url'

            task.driver.boot.prepare_ramdisk(task, {})

            mock_node_power_action.assert_called_once_with(
                task, states.POWER_OFF)

            mock__eject_vmedia.assert_called_once_with(
                task, managers, sushy.VIRTUAL_MEDIA_CD)

            mock__insert_vmedia.assert_called_once_with(
                task, managers, 'image-url', sushy.VIRTUAL_MEDIA_CD)

            expected_params = {
                'ipa-agent-token': mock.ANY,
                'boot_method': 'vmedia',
            }

            mock_prepare_deploy_iso.assert_called_once_with(
                task, expected_params, 'deploy', {})

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.CDROM, False)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)

    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(image_utils, 'prepare_floppy_image', autospec=True)
    @mock.patch.object(image_utils, 'prepare_deploy_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_has_vmedia_device', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_insert_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_driver_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_power_action',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_ramdisk_with_floppy(
            self, mock_system, mock_boot_mode_utils, mock_node_power_action,
            mock__parse_driver_info, mock__insert_vmedia, mock__eject_vmedia,
            mock__has_vmedia_device, mock_prepare_deploy_iso,
            mock_prepare_floppy_image, mock_node_set_boot_device):
        system = mock_system.return_value
        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.node.provision_state = states.DEPLOYING

            d_info = {
                'config_via_removable': True
            }

            mock__parse_driver_info.return_value = d_info

            mock__has_vmedia_device.return_value = sushy.VIRTUAL_MEDIA_FLOPPY
            mock_prepare_floppy_image.return_value = 'floppy-image-url'
            mock_prepare_deploy_iso.return_value = 'cd-image-url'

            task.driver.boot.prepare_ramdisk(task, {})

            mock_node_power_action.assert_called_once_with(
                task, states.POWER_OFF)

            mock__has_vmedia_device.assert_called_once_with(
                managers,
                [sushy.VIRTUAL_MEDIA_USBSTICK, sushy.VIRTUAL_MEDIA_FLOPPY],
                system=system)

            eject_calls = [
                mock.call(task, managers, dev)
                for dev in (sushy.VIRTUAL_MEDIA_FLOPPY,
                            sushy.VIRTUAL_MEDIA_CD)
            ]

            mock__eject_vmedia.assert_has_calls(eject_calls)

            insert_calls = [
                mock.call(task, managers, 'floppy-image-url',
                          sushy.VIRTUAL_MEDIA_FLOPPY),
                mock.call(task, managers, 'cd-image-url',
                          sushy.VIRTUAL_MEDIA_CD),
            ]

            mock__insert_vmedia.assert_has_calls(insert_calls)

            expected_params = {
                'boot_method': 'vmedia',
                'ipa-debug': '1',
                'ipa-agent-token': mock.ANY,
            }

            mock_prepare_deploy_iso.assert_called_once_with(
                task, expected_params, 'deploy', d_info)

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.CDROM, False)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)

    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(image_utils, 'prepare_floppy_image', autospec=True)
    @mock.patch.object(image_utils, 'prepare_deploy_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_has_vmedia_device', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_insert_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_driver_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_power_action',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_ramdisk_with_usb(
            self, mock_system, mock_boot_mode_utils, mock_node_power_action,
            mock__parse_driver_info, mock__insert_vmedia, mock__eject_vmedia,
            mock__has_vmedia_device, mock_prepare_deploy_iso,
            mock_prepare_floppy_image, mock_node_set_boot_device):
        system = mock_system.return_value
        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.node.provision_state = states.DEPLOYING

            d_info = {
                'config_via_removable': True
            }

            mock__parse_driver_info.return_value = d_info

            mock__has_vmedia_device.return_value = sushy.VIRTUAL_MEDIA_USBSTICK
            mock_prepare_floppy_image.return_value = 'floppy-image-url'
            mock_prepare_deploy_iso.return_value = 'cd-image-url'

            task.driver.boot.prepare_ramdisk(task, {})

            mock_node_power_action.assert_called_once_with(
                task, states.POWER_OFF)

            mock__has_vmedia_device.assert_called_once_with(
                managers,
                [sushy.VIRTUAL_MEDIA_USBSTICK, sushy.VIRTUAL_MEDIA_FLOPPY],
                system=system)

            eject_calls = [
                mock.call(task, managers, dev)
                for dev in (sushy.VIRTUAL_MEDIA_USBSTICK,
                            sushy.VIRTUAL_MEDIA_CD)
            ]

            mock__eject_vmedia.assert_has_calls(eject_calls)

            insert_calls = [
                mock.call(task, managers, 'floppy-image-url',
                          sushy.VIRTUAL_MEDIA_USBSTICK),
                mock.call(task, managers, 'cd-image-url',
                          sushy.VIRTUAL_MEDIA_CD),
            ]

            mock__insert_vmedia.assert_has_calls(insert_calls)

            expected_params = {
                'boot_method': 'vmedia',
                'ipa-debug': '1',
                'ipa-agent-token': mock.ANY,
            }

            mock_prepare_deploy_iso.assert_called_once_with(
                task, expected_params, 'deploy', d_info)

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.CDROM, False)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)

    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(image_utils, 'prepare_deploy_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_insert_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_driver_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_power_action',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_ramdisk_no_config(
            self, mock_system, mock_boot_mode_utils, mock_node_power_action,
            mock__parse_driver_info, mock__insert_vmedia, mock__eject_vmedia,
            mock_prepare_deploy_iso, mock_node_set_boot_device):

        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.node.provision_state = states.DEPLOYING

            mock__parse_driver_info.return_value = {
                'can_provide_config': False}
            mock_prepare_deploy_iso.return_value = 'image-url'

            task.driver.boot.prepare_ramdisk(task, {})

            mock_node_power_action.assert_called_once_with(
                task, states.POWER_OFF)

            mock__eject_vmedia.assert_called_once_with(
                task, managers, sushy.VIRTUAL_MEDIA_CD)

            mock__insert_vmedia.assert_called_once_with(
                task, managers, 'image-url', sushy.VIRTUAL_MEDIA_CD)

            expected_params = {
                'ipa-debug': '1',
                'boot_method': 'vmedia',
            }

            mock_prepare_deploy_iso.assert_called_once_with(
                task, expected_params, 'deploy', {})

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.CDROM, False)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)

            self.assertNotIn('agent_secret_token',
                             task.node.driver_internal_info)
            self.assertNotIn('agent_secret_token_pregenerated',
                             task.node.driver_internal_info)

    @mock.patch.object(manager_utils, 'is_fast_track', lambda task: True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(image_utils, 'prepare_deploy_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_has_vmedia_device', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_insert_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_driver_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_power_action',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_ramdisk_fast_track(
            self, mock_system, mock_boot_mode_utils, mock_node_power_action,
            mock__parse_driver_info, mock__insert_vmedia, mock__eject_vmedia,
            mock__has_vmedia_device,
            mock_prepare_deploy_iso, mock_node_set_boot_device):
        system = mock_system.return_value
        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.node.provision_state = states.DEPLOYING
            mock__has_vmedia_device.return_value = sushy.VIRTUAL_MEDIA_CD

            task.driver.boot.prepare_ramdisk(task, {})

            mock__has_vmedia_device.assert_called_once_with(
                managers, sushy.VIRTUAL_MEDIA_CD, inserted=True,
                system=system)

            mock_node_power_action.assert_not_called()
            mock__eject_vmedia.assert_not_called()
            mock__insert_vmedia.assert_not_called()
            mock_prepare_deploy_iso.assert_not_called()
            mock_node_set_boot_device.assert_not_called()
            mock_boot_mode_utils.sync_boot_mode.assert_not_called()

    @mock.patch.object(manager_utils, 'is_fast_track', lambda task: True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(image_utils, 'prepare_deploy_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_has_vmedia_device', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_insert_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_driver_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_power_action',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_ramdisk_fast_track_impossible(
            self, mock_system, mock_boot_mode_utils, mock_node_power_action,
            mock__parse_driver_info, mock__insert_vmedia, mock__eject_vmedia,
            mock__has_vmedia_device,
            mock_prepare_deploy_iso, mock_node_set_boot_device):
        system = mock_system.return_value
        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.node.provision_state = states.DEPLOYING
            mock__has_vmedia_device.return_value = False

            mock__parse_driver_info.return_value = {}
            mock_prepare_deploy_iso.return_value = 'image-url'

            task.driver.boot.prepare_ramdisk(task, {})

            mock__has_vmedia_device.assert_called_once_with(
                managers, sushy.VIRTUAL_MEDIA_CD, inserted=True,
                system=system)

            mock_node_power_action.assert_called_once_with(
                task, states.POWER_OFF)

            mock__eject_vmedia.assert_called_once_with(
                task, managers, sushy.VIRTUAL_MEDIA_CD)

            mock__insert_vmedia.assert_called_once_with(
                task, managers, 'image-url', sushy.VIRTUAL_MEDIA_CD)

            token = task.node.driver_internal_info['agent_secret_token']
            self.assertTrue(token)

            expected_params = {
                'ipa-agent-token': token,
                'ipa-debug': '1',
                'boot_method': 'vmedia',
            }

            mock_prepare_deploy_iso.assert_called_once_with(
                task, expected_params, 'deploy', {})

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.CDROM, False)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)

            self.assertTrue(task.node.driver_internal_info[
                'agent_secret_token_pregenerated'])

    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_floppy_image', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_driver_info', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_clean_up_ramdisk(
            self, mock_system, mock__parse_driver_info,
            mock_cleanup_floppy_image, mock_cleanup_iso_image,
            mock__eject_vmedia):

        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            task.node.driver_info['config_via_removable'] = True

            task.driver.boot.clean_up_ramdisk(task)

            mock_cleanup_iso_image.assert_called_once_with(task)

            mock_cleanup_floppy_image.assert_called_once_with(task)

            eject_calls = [
                mock.call(task, managers, sushy.VIRTUAL_MEDIA_CD),
                mock.call(task, managers, sushy.VIRTUAL_MEDIA_USBSTICK),
                mock.call(task, managers, sushy.VIRTUAL_MEDIA_FLOPPY)
            ]

            mock__eject_vmedia.assert_has_calls(eject_calls)

    @mock.patch.object(redfish_boot.RedfishVirtualMediaBoot,
                       '_eject_all', autospec=True)
    @mock.patch.object(image_utils, 'prepare_boot_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_insert_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_deploy_info', autospec=True)
    @mock.patch.object(redfish_boot, 'manager_utils', autospec=True)
    @mock.patch.object(redfish_boot, 'deploy_utils', autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_instance_normal_boot(
            self, mock_system, mock_boot_mode_utils, mock_deploy_utils,
            mock_manager_utils, mock__parse_deploy_info, mock__insert_vmedia,
            mock__eject_vmedia, mock_prepare_boot_iso, mock_clean_up_instance):

        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            task.node.driver_internal_info[
                'root_uuid_or_disk_id'] = self.node.uuid

            mock_deploy_utils.get_boot_option.return_value = 'net'

            d_info = {
                'deploy_kernel': 'kernel',
                'deploy_ramdisk': 'ramdisk',
                'bootloader': 'bootloader'
            }

            mock__parse_deploy_info.return_value = d_info

            mock_prepare_boot_iso.return_value = 'image-url'

            task.driver.boot.prepare_instance(task)

            expected_params = {
                'root_uuid': self.node.uuid
            }

            mock_prepare_boot_iso.assert_called_once_with(
                task, d_info, **expected_params)

            mock__eject_vmedia.assert_called_once_with(
                task, managers, sushy.VIRTUAL_MEDIA_CD)

            mock__insert_vmedia.assert_called_once_with(
                task, managers, 'image-url', sushy.VIRTUAL_MEDIA_CD)

            mock_manager_utils.node_set_boot_device.assert_called_once_with(
                task, boot_devices.CDROM, persistent=True)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)
            csb = mock_boot_mode_utils.configure_secure_boot_if_needed
            csb.assert_called_once_with(task)

    @mock.patch.object(redfish_boot.RedfishVirtualMediaBoot,
                       '_eject_all', autospec=True)
    @mock.patch.object(image_utils, 'prepare_configdrive_image', autospec=True)
    @mock.patch.object(image_utils, 'prepare_boot_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_insert_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_deploy_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'deploy_utils', autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_instance_ramdisk_boot(
            self, mock_system, mock_boot_mode_utils, mock_deploy_utils,
            mock_node_set_boot_device, mock__parse_deploy_info,
            mock__insert_vmedia, mock__eject_vmedia, mock_prepare_boot_iso,
            mock_prepare_disk, mock_clean_up_instance):

        configdrive = 'Y29udGVudA=='
        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            task.node.driver_internal_info[
                'root_uuid_or_disk_id'] = self.node.uuid
            task.node.instance_info['configdrive'] = configdrive

            mock_deploy_utils.get_boot_option.return_value = 'ramdisk'

            d_info = {
                'deploy_kernel': 'kernel',
                'deploy_ramdisk': 'ramdisk',
                'bootloader': 'bootloader'
            }
            mock__parse_deploy_info.return_value = d_info

            mock_prepare_boot_iso.return_value = 'image-url'
            mock_prepare_disk.return_value = 'cd-url'

            task.driver.boot.prepare_instance(task)

            mock_clean_up_instance.assert_called_once_with(mock.ANY, task)

            mock_prepare_boot_iso.assert_called_once_with(task, d_info)
            mock_prepare_disk.assert_called_once_with(task, configdrive)

            mock__eject_vmedia.assert_has_calls([
                mock.call(task, managers, sushy.VIRTUAL_MEDIA_CD),
                mock.call(task, managers, sushy.VIRTUAL_MEDIA_USBSTICK),
            ])

            mock__insert_vmedia.assert_has_calls([
                mock.call(task, managers,
                          'image-url', sushy.VIRTUAL_MEDIA_CD),
                mock.call(task, managers,
                          'cd-url', sushy.VIRTUAL_MEDIA_USBSTICK),
            ])

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.CDROM, persistent=True)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)

    @mock.patch.object(redfish_boot.RedfishVirtualMediaBoot,
                       '_eject_all', autospec=True)
    @mock.patch.object(image_utils, 'prepare_boot_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_insert_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_deploy_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'deploy_utils', autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_instance_ramdisk_boot_iso(
            self, mock_system, mock_boot_mode_utils, mock_deploy_utils,
            mock_node_set_boot_device, mock__parse_deploy_info,
            mock__insert_vmedia, mock__eject_vmedia, mock_prepare_boot_iso,
            mock_clean_up_instance):

        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            task.node.driver_internal_info[
                'root_uuid_or_disk_id'] = self.node.uuid
            task.node.instance_info['configdrive'] = None

            mock_deploy_utils.get_boot_option.return_value = 'ramdisk'

            d_info = {
                'deploy_kernel': 'kernel',
                'deploy_ramdisk': 'ramdisk',
                'bootloader': 'bootloader'
            }

            mock__parse_deploy_info.return_value = d_info
            mock_prepare_boot_iso.return_value = 'image-url'

            task.driver.boot.prepare_instance(task)

            mock_prepare_boot_iso.assert_called_once_with(task, d_info)

            mock__eject_vmedia.assert_called_once_with(
                task, managers, sushy.VIRTUAL_MEDIA_CD)

            mock__insert_vmedia.assert_called_once_with(
                task, managers, 'image-url', sushy.VIRTUAL_MEDIA_CD)

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.CDROM, persistent=True)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)

    @mock.patch.object(redfish_boot.RedfishVirtualMediaBoot,
                       '_eject_all', autospec=True)
    @mock.patch.object(image_utils, 'prepare_boot_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_insert_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_deploy_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'deploy_utils', autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_instance_ramdisk_boot_iso_boot(
            self, mock_system, mock_boot_mode_utils, mock_deploy_utils,
            mock_node_set_boot_device, mock__parse_deploy_info,
            mock__insert_vmedia, mock__eject_vmedia, mock_prepare_boot_iso,
            mock_clean_up_instance):

        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            i_info = task.node.instance_info
            i_info['boot_iso'] = "super-magic"
            del i_info['configdrive']
            task.node.instance_info = i_info
            mock_deploy_utils.get_boot_option.return_value = 'ramdisk'
            mock__parse_deploy_info.return_value = {}

            mock_prepare_boot_iso.return_value = 'image-url'

            task.driver.boot.prepare_instance(task)

            mock_prepare_boot_iso.assert_called_once_with(task, {})

            mock__eject_vmedia.assert_called_once_with(
                task, managers, sushy.VIRTUAL_MEDIA_CD)

            mock__insert_vmedia.assert_called_once_with(
                task, managers, 'image-url', sushy.VIRTUAL_MEDIA_CD)

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.CDROM, persistent=True)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)

    @mock.patch.object(redfish_boot.manager_utils, 'build_configdrive',
                       autospec=True)
    @mock.patch.object(redfish_boot.RedfishVirtualMediaBoot,
                       '_eject_all', autospec=True)
    @mock.patch.object(image_utils, 'prepare_configdrive_image', autospec=True)
    @mock.patch.object(image_utils, 'prepare_boot_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_insert_vmedia', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_deploy_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'deploy_utils', autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_instance_ramdisk_boot_render_configdrive(
            self, mock_system, mock_boot_mode_utils, mock_deploy_utils,
            mock_node_set_boot_device, mock__parse_deploy_info,
            mock__insert_vmedia, mock__eject_vmedia, mock_prepare_boot_iso,
            mock_prepare_disk, mock_clean_up_instance, mock_build_configdrive):

        configdrive = 'Y29udGVudA=='
        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            task.node.driver_internal_info[
                'root_uuid_or_disk_id'] = self.node.uuid
            task.node.instance_info['configdrive'] = {'meta_data': {}}

            mock_build_configdrive.return_value = configdrive

            mock_deploy_utils.get_boot_option.return_value = 'ramdisk'

            d_info = {
                'deploy_kernel': 'kernel',
                'deploy_ramdisk': 'ramdisk',
                'bootloader': 'bootloader'
            }
            mock__parse_deploy_info.return_value = d_info

            mock_prepare_boot_iso.return_value = 'image-url'
            mock_prepare_disk.return_value = 'cd-url'

            task.driver.boot.prepare_instance(task)

            mock_clean_up_instance.assert_called_once_with(mock.ANY, task)

            mock_build_configdrive.assert_called_once_with(
                task.node, {'meta_data': {}})
            mock_prepare_boot_iso.assert_called_once_with(task, d_info)
            mock_prepare_disk.assert_called_once_with(task, configdrive)

            mock__eject_vmedia.assert_has_calls([
                mock.call(task, managers, sushy.VIRTUAL_MEDIA_CD),
                mock.call(task, managers, sushy.VIRTUAL_MEDIA_USBSTICK),
            ])

            mock__insert_vmedia.assert_has_calls([
                mock.call(task, managers,
                          'image-url', sushy.VIRTUAL_MEDIA_CD),
                mock.call(task, managers,
                          'cd-url', sushy.VIRTUAL_MEDIA_USBSTICK),
            ])

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.CDROM, persistent=True)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)

    @mock.patch.object(boot_mode_utils, 'configure_secure_boot_if_needed',
                       autospec=True)
    @mock.patch.object(boot_mode_utils, 'sync_boot_mode', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(redfish_boot, 'manager_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def _test_prepare_instance_local_boot(
            self, mock_system, mock_manager_utils,
            mock_cleanup_iso_image, mock__eject_vmedia, mock_sync_boot_mode,
            mock_secure_boot):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            task.node.driver_internal_info[
                'root_uuid_or_disk_id'] = self.node.uuid

            task.driver.boot.prepare_instance(task)

            mock_manager_utils.node_set_boot_device.assert_called_once_with(
                task, boot_devices.DISK, persistent=True)
            mock_cleanup_iso_image.assert_called_once_with(task)
            mock__eject_vmedia.assert_called_once_with(
                task, mock_system.return_value.managers,
                sushy.VIRTUAL_MEDIA_CD)
            mock_sync_boot_mode.assert_called_once_with(task)
            mock_secure_boot.assert_called_once_with(task)

    def test_prepare_instance_local_whole_disk_image(self):
        self.node.driver_internal_info = {'is_whole_disk_image': True}
        self.node.save()
        self._test_prepare_instance_local_boot()

    def test_prepare_instance_local_boot_option(self):
        instance_info = self.node.instance_info
        instance_info['capabilities'] = '{"boot_option": "local"}'
        self.node.instance_info = instance_info
        self.node.save()
        self._test_prepare_instance_local_boot()

    @mock.patch.object(boot_mode_utils, 'deconfigure_secure_boot_if_needed',
                       autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_floppy_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def _test_clean_up_instance(self, mock_system, mock_cleanup_iso_image,
                                mock_cleanup_floppy_image,
                                mock__eject_vmedia, mock_secure_boot):
        managers = mock_system.return_value.managers
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:

            task.driver.boot.clean_up_instance(task)

            mock_cleanup_iso_image.assert_called_once_with(task)
            eject_calls = [mock.call(task, managers, sushy.VIRTUAL_MEDIA_CD)]
            if task.node.driver_info.get('config_via_removable'):
                eject_calls.extend([
                    mock.call(task, managers, sushy.VIRTUAL_MEDIA_USBSTICK),
                    mock.call(task, managers, sushy.VIRTUAL_MEDIA_FLOPPY),
                ])
                mock_cleanup_floppy_image.assert_called_once_with(task)

            mock__eject_vmedia.assert_has_calls(eject_calls)
            mock_secure_boot.assert_called_once_with(task)

    def test_clean_up_instance_only_cdrom(self):
        self._test_clean_up_instance()

    def test_clean_up_instance_cdrom_and_floppy(self):
        driver_info = self.node.driver_info
        driver_info['config_via_removable'] = True
        self.node.driver_info = driver_info
        self.node.save()
        self._test_clean_up_instance()

    @mock.patch.object(boot_mode_utils, 'deconfigure_secure_boot_if_needed',
                       autospec=True)
    @mock.patch.object(deploy_utils, 'get_boot_option', autospec=True)
    @mock.patch.object(redfish_boot, '_eject_vmedia', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_disk_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_clean_up_instance_ramdisk(self, mock_system,
                                       mock_cleanup_iso_image,
                                       mock_cleanup_disk_image,
                                       mock__eject_vmedia,
                                       mock_get_boot_option,
                                       mock_secure_boot):
        managers = mock_system.return_value.managers

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_get_boot_option.return_value = 'ramdisk'

            task.driver.boot.clean_up_instance(task)

            mock_cleanup_iso_image.assert_called_once_with(task)
            mock_cleanup_disk_image.assert_called_once_with(
                task, prefix='configdrive')
            eject_calls = [
                mock.call(task, managers, sushy.VIRTUAL_MEDIA_CD),
                mock.call(task, managers, sushy.VIRTUAL_MEDIA_USBSTICK),
            ]

            mock__eject_vmedia.assert_has_calls(eject_calls)
            mock_secure_boot.assert_called_once_with(task)

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__insert_vmedia_anew(self, mock_sys, mock_vmd_sys):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmd_sys.return_value = False
            mock_vmedia_cd = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_CD])
            mock_vmedia_floppy = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])

            mock_manager = mock.MagicMock()

            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_cd, mock_vmedia_floppy]

            redfish_boot._insert_vmedia(
                task, [mock_manager], 'img-url', sushy.VIRTUAL_MEDIA_CD)

            mock_vmedia_cd.insert_media.assert_called_once_with(
                'img-url', inserted=True, write_protected=True)

            self.assertFalse(mock_vmedia_floppy.insert_media.call_count)

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__insert_vmedia_anew_dvd(self, mock_sys, mock_vmd_sys):
        mock_vmd_sys.return_value = False
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmedia_dvd = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_DVD])

            mock_manager = mock.MagicMock()

            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_dvd]

            redfish_boot._insert_vmedia(
                task, [mock_manager], 'img-url', sushy.VIRTUAL_MEDIA_CD)

            mock_vmedia_dvd.insert_media.assert_called_once_with(
                'img-url', inserted=True, write_protected=True)

    @mock.patch('time.sleep', lambda *args, **kwargs: None)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__insert_vmedia_anew_dvd_retry(self, mock_sys, mock_vmd_sys):
        mock_vmd_sys.return_value = False
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmedia_dvd_1 = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_DVD])

            mock_vmedia_dvd_2 = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_DVD])

            mock_manager = mock.MagicMock()

            def clear_and_raise(*args, **kwargs):
                mock_vmedia_dvd_1.insert_media.side_effect = None
                raise sushy.exceptions.BadRequestError(
                    "POST", 'img-url', mock.MagicMock())
            mock_vmedia_dvd_1.insert_media.side_effect = clear_and_raise
            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_dvd_1, mock_vmedia_dvd_2]

            redfish_boot._insert_vmedia(
                task, [mock_manager], 'img-url', sushy.VIRTUAL_MEDIA_CD)

            self.assertEqual(mock_vmedia_dvd_2.insert_media.call_count, 1)

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__insert_vmedia_already_inserted(self, mock_sys, mock_vmd_sys):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmd_sys.return_value = False
            mock_vmedia_cd = mock.MagicMock(
                inserted=True,
                image='img-url',
                media_types=[sushy.VIRTUAL_MEDIA_CD])
            mock_manager = mock.MagicMock()

            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_cd]

            redfish_boot._insert_vmedia(
                task, [mock_manager], 'img-url', sushy.VIRTUAL_MEDIA_CD)

            self.assertFalse(mock_vmedia_cd.insert_media.call_count)

    @mock.patch('time.sleep', lambda *args, **kwargs: None)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__insert_vmedia_while_ejecting(self, mock_sys, mock_vmd_sys):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmd_sys.return_value = False
            mock_vmedia_cd = mock.MagicMock(
                inserted=False,
                image='img-url',
                media_types=[sushy.VIRTUAL_MEDIA_CD],
            )
            mock_manager = mock.MagicMock()

            def clear_and_raise(*args, **kwargs):
                mock_vmedia_cd.insert_media.side_effect = None
                raise sushy.exceptions.ServerSideError(
                    "POST", 'img-url', mock.MagicMock())
            mock_vmedia_cd.insert_media.side_effect = clear_and_raise
            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_cd]

            redfish_boot._insert_vmedia(
                task, [mock_manager], 'img-url', sushy.VIRTUAL_MEDIA_CD)

            self.assertEqual(mock_vmedia_cd.insert_media.call_count, 2)

    @mock.patch('time.sleep', lambda *args, **kwargs: None)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__insert_vmedia_bad_device(self, mock_sys, mock_vmd_sys):
        mock_vmd_sys.return_value = False
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmedia_floppy = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])
            mock_manager = mock.MagicMock()

            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_floppy]

            self.assertRaises(
                exception.InvalidParameterValue,
                redfish_boot._insert_vmedia,
                task, [mock_manager], 'img-url', sushy.VIRTUAL_MEDIA_CD)

    @mock.patch('time.sleep', lambda *args, **kwargs: None)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__insert_vmedia_empty_media_type(self, mock_sys, mock_vmd_sys):
        mock_vmd_sys.return_value = False
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmedia_empty_media_type = mock.MagicMock(
                inserted=False,
                media_types=[])
            mock_manager = mock.MagicMock()

            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_empty_media_type]

            self.assertRaises(
                exception.InvalidParameterValue,
                redfish_boot._insert_vmedia,
                task, [mock_manager], 'img-url', sushy.VIRTUAL_MEDIA_CD)

    @mock.patch.object(image_utils, 'cleanup_disk_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_boot, 'redfish_utils', autospec=True)
    def test_eject_vmedia_everything(self, mock_redfish_utils, mock_vmd_sys,
                                     mock_cleanup_iso, mock_cleanup_disk):
        mock_vmd_sys.return_value = False
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmedia_cd = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_CD])
            mock_vmedia_floppy = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])
            mock_vmedia_dvd = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_DVD])

            mock_manager = mock.MagicMock()

            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_cd, mock_vmedia_floppy, mock_vmedia_dvd]

            mock_redfish_utils.get_system.return_value.managers = [
                mock_manager]

            redfish_boot.eject_vmedia(task)

            mock_vmedia_cd.eject_media.assert_called_once_with()
            mock_vmedia_floppy.eject_media.assert_called_once_with()
            mock_vmedia_dvd.eject_media.assert_called_once_with()
            mock_cleanup_iso.assert_called_once_with(task)
            mock_cleanup_disk.assert_called_once_with(task,
                                                      prefix='configdrive')

    @mock.patch.object(image_utils, 'cleanup_disk_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_boot, 'redfish_utils', autospec=True)
    def test_eject_vmedia_specific(self, mock_redfish_utils, mock_vmd_sys,
                                   mock_cleanup_iso, mock_cleanup_disk):
        mock_vmd_sys.return_value = False
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmedia_cd = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_CD])
            mock_vmedia_floppy = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])

            mock_manager = mock.MagicMock()

            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_cd, mock_vmedia_floppy]

            mock_redfish_utils.get_system.return_value.managers = [
                mock_manager]

            redfish_boot.eject_vmedia(task, sushy.VIRTUAL_MEDIA_CD)

            mock_vmedia_cd.eject_media.assert_called_once_with()
            self.assertFalse(mock_vmedia_floppy.eject_media.call_count)
            mock_cleanup_iso.assert_called_once_with(task)
            mock_cleanup_disk.assert_not_called()

    @mock.patch.object(image_utils, 'cleanup_disk_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(redfish_boot, 'redfish_utils', autospec=True)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_boot.LOG, 'debug', autospec=True)
    @mock.patch.object(redfish_boot.LOG, 'info', autospec=True)
    def test_eject_vmedia_with_dvd_cisco_ucs(self, mock_log_info,
                                             mock_log_debug,
                                             mock_vmd_sys,
                                             mock_redfish_utils,
                                             mock_cleanup_iso,
                                             mock_cleanup_disk):
        mock_vmd_sys.return_value = False
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmedia_dvd_1 = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_DVD])
            mock_vmedia_dvd_2 = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_DVD])

            mock_manager = mock.MagicMock()

            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_dvd_1, mock_vmedia_dvd_2]

            mock_redfish_utils.get_system.return_value.managers = [
                mock_manager]

            redfish_boot.eject_vmedia(task, sushy.VIRTUAL_MEDIA_CD)

            mock_vmedia_dvd_1.eject_media.assert_called_once_with()
            mock_vmedia_dvd_2.eject_media.assert_called_once_with()

            self.assertEqual(mock_log_info.call_count, 2)
            self.assertEqual(mock_log_debug.call_count, 3)
            mock_cleanup_iso.assert_called_once_with(task)
            mock_cleanup_disk.assert_not_called()

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_boot, 'redfish_utils', autospec=True)
    def test_eject_vmedia_not_inserted(self, mock_redfish_utils, mock_vmd_sys):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmd_sys.return_value = False
            mock_vmedia_cd = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_CD])
            mock_vmedia_floppy = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])

            mock_manager = mock.MagicMock()

            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_cd, mock_vmedia_floppy]

            mock_redfish_utils.get_system.return_value.managers = [
                mock_manager]

            redfish_boot.eject_vmedia(task)

            self.assertFalse(mock_vmedia_cd.eject_media.call_count)
            self.assertFalse(mock_vmedia_floppy.eject_media.call_count)

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_boot, 'redfish_utils', autospec=True)
    def test_eject_vmedia_unknown(self, mock_redfish_utils, mock_vmd_sys):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmd_sys.return_value = False
            mock_vmedia_cd = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_CD])

            mock_manager = mock.MagicMock()

            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_cd]

            mock_redfish_utils.get_system.return_value.managers = [
                mock_manager]

            redfish_boot.eject_vmedia(task)

            self.assertFalse(mock_vmedia_cd.eject_media.call_count)

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    def test__has_vmedia_device(self, mock_vmd_sys):
        mock_vmd_sys.return_value = False
        mock_vmedia_cd = mock.MagicMock(
            inserted=False,
            media_types=[sushy.VIRTUAL_MEDIA_CD])
        mock_vmedia_floppy = mock.MagicMock(
            inserted=False,
            media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])

        mock_manager = mock.MagicMock()

        mock_manager.virtual_media.get_members.return_value = [
            mock_vmedia_cd, mock_vmedia_floppy]

        self.assertEqual(
            sushy.VIRTUAL_MEDIA_CD,
            redfish_boot._has_vmedia_device(
                [mock_manager], sushy.VIRTUAL_MEDIA_CD))

        self.assertFalse(
            redfish_boot._has_vmedia_device(
                [mock_manager], sushy.VIRTUAL_MEDIA_CD, inserted=True))

        self.assertFalse(
            redfish_boot._has_vmedia_device(
                [mock_manager], sushy.VIRTUAL_MEDIA_USBSTICK))

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    def test__has_vmedia_device_inserted(self, mock_vmd_sys):
        mock_vmd_sys.return_value = False
        mock_vmedia_cd = mock.MagicMock(
            inserted=False,
            media_types=[sushy.VIRTUAL_MEDIA_CD])
        mock_vmedia_floppy = mock.MagicMock(
            inserted=True,
            media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])

        mock_manager = mock.MagicMock()

        mock_manager.virtual_media.get_members.return_value = [
            mock_vmedia_cd, mock_vmedia_floppy]

        self.assertEqual(
            sushy.VIRTUAL_MEDIA_FLOPPY,
            redfish_boot._has_vmedia_device(
                [mock_manager], sushy.VIRTUAL_MEDIA_FLOPPY, inserted=True))


@mock.patch('oslo_utils.eventletutils.EventletEvent.wait',
            lambda *args, **kwargs: None)
class RedfishHTTPBootTestCase(db_base.DbTestCase):

    def setUp(self):
        super(RedfishHTTPBootTestCase, self).setUp()
        self.config(enabled_hardware_types=['redfish'],
                    enabled_power_interfaces=['redfish'],
                    enabled_boot_interfaces=['redfish-https'],
                    enabled_management_interfaces=['redfish'],
                    enabled_inspect_interfaces=['redfish'],
                    enabled_bios_interfaces=['redfish'])
        self.node = obj_utils.create_test_node(
            self.context, driver='redfish', driver_info=INFO_DICT)

    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(image_utils, 'prepare_deploy_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_driver_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_power_action',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_ramdisk_with_params(
            self, mock_system, mock_boot_mode_utils, mock_node_power_action,
            mock__parse_driver_info,
            mock_prepare_deploy_iso, mock_node_set_boot_device):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.node.provision_state = states.DEPLOYING

            mock__parse_driver_info.return_value = {}
            mock_prepare_deploy_iso.return_value = 'image-url'

            task.driver.boot.prepare_ramdisk(task, {})

            mock_node_power_action.assert_called_once_with(
                task, states.POWER_OFF)

            token = task.node.driver_internal_info['agent_secret_token']
            self.assertTrue(token)

            expected_params = {
                'ipa-agent-token': token,
                'ipa-debug': '1',
                'boot_method': 'vmedia',
            }

            mock_prepare_deploy_iso.assert_called_once_with(
                task, expected_params, 'deploy', {})

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.UEFIHTTP, False)
            self.assertEqual('image-url',
                             task.node.driver_internal_info.get(
                                 'redfish_uefi_http_url'))
            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)
            self.assertTrue(task.node.driver_internal_info[
                'agent_secret_token_pregenerated'])

    @mock.patch.object(deploy_utils, 'get_boot_option', lambda node: 'ramdisk')
    def test_parse_driver_info_ramdisk(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.driver_info = {}
            task.node.automated_clean = False
            actual_driver_info = redfish_boot._parse_driver_info(task.node)
            self.assertEqual({'can_provide_config': False},
                             actual_driver_info)

    def test_parse_driver_info_deploy(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            actual_driver_info = redfish_boot._parse_driver_info(task.node)

            self.assertIn('kernel', actual_driver_info['deploy_kernel'])
            self.assertIn('ramdisk', actual_driver_info['deploy_ramdisk'])
            self.assertIn('bootloader', actual_driver_info['bootloader'])
            self.assertTrue(actual_driver_info['can_provide_config'])

    def test_parse_driver_info_iso(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.driver_info.update(
                {'deploy_iso': 'http://boot.iso'})

            actual_driver_info = redfish_boot._parse_driver_info(task.node)

            self.assertEqual('http://boot.iso',
                             actual_driver_info['deploy_iso'])
            self.assertFalse(actual_driver_info['can_provide_config'])

    def test_parse_driver_info_rescue(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.RESCUING
            task.node.driver_info.update(
                {'rescue_kernel': 'kernel',
                 'rescue_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            actual_driver_info = redfish_boot._parse_driver_info(task.node)

            self.assertIn('kernel', actual_driver_info['rescue_kernel'])
            self.assertIn('ramdisk', actual_driver_info['rescue_ramdisk'])
            self.assertIn('bootloader', actual_driver_info['bootloader'])

    def test_parse_driver_info_exc(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.MissingParameterValue,
                              redfish_boot._parse_driver_info,
                              task.node)

    def _test_parse_driver_info_from_conf(self, mode='deploy', by_arch=False):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            if mode == 'rescue':
                task.node.provision_state = states.RESCUING

            if by_arch:
                ramdisk = 'glance://%s_ramdisk_uuid' % mode
                kernel = 'glance://%s_kernel_uuid' % mode

                config = {
                    '%s_ramdisk_by_arch' % mode: {'x86_64': ramdisk},
                    '%s_kernel_by_arch' % mode: {'x86_64': kernel}
                }
                expected = {
                    '%s_ramdisk' % mode: ramdisk,
                    '%s_kernel' % mode: kernel
                }
            else:
                expected = {
                    '%s_ramdisk' % mode: 'glance://%s_ramdisk_uuid' % mode,
                    '%s_kernel' % mode: 'glance://%s_kernel_uuid' % mode
                }
                config = expected

            self.config(group='conductor', **config)

            image_info = redfish_boot._parse_driver_info(task.node)

            for key, value in expected.items():
                self.assertEqual(value, image_info[key])

    def test_parse_driver_info_from_conf_deploy(self):
        self._test_parse_driver_info_from_conf()

    def test_parse_driver_info_from_conf_rescue(self):
        self._test_parse_driver_info_from_conf(mode='rescue')

    def test_parse_driver_info_from_conf_deploy_by_arch(self):
        self._test_parse_driver_info_from_conf(by_arch=True)

    def test_parse_driver_info_from_conf_rescue_by_arch(self):
        self._test_parse_driver_info_from_conf(mode='rescue', by_arch=True)

    def _test_parse_driver_info_mixed_source(self, mode='deploy',
                                             by_arch=False):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            if mode == 'rescue':
                task.node.provision_state = states.RESCUING

            if by_arch:
                kernel_config = {
                    '%s_kernel_by_arch' % mode: {
                        'x86': 'glance://%s_kernel_uuid' % mode
                    }
                }
            else:
                kernel_config = {
                    '%s_kernel' % mode: 'glance://%s_kernel_uuid' % mode
                }

            ramdisk_config = {
                '%s_ramdisk' % mode: 'glance://%s_ramdisk_uuid' % mode,
            }

            self.config(group='conductor', **kernel_config)

            task.node.driver_info.update(ramdisk_config)

            self.assertRaises(exception.MissingParameterValue,
                              redfish_boot._parse_driver_info, task.node)

    def test_parse_driver_info_mixed_source_deploy(self):
        self._test_parse_driver_info_mixed_source()

    def test_parse_driver_info_mixed_source_rescue(self):
        self._test_parse_driver_info_mixed_source(mode='rescue')

    def test_parse_driver_info_mixed_source_deploy_by_arch(self):
        self._test_parse_driver_info_mixed_source(by_arch=True)

    def test_parse_driver_info_mixed_source_rescue_by_arch(self):
        self._test_parse_driver_info_mixed_source(mode='rescue', by_arch=True)

    def _test_parse_driver_info_choose_by_arch(self, mode='deploy'):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            if mode == 'rescue':
                task.node.provision_state = states.RESCUING
            task.node.properties['cpu_arch'] = 'aarch64'
            wrong_ramdisk = 'glance://wrong_%s_ramdisk_uuid' % mode
            wrong_kernel = 'glance://wrong_%s_kernel_uuid' % mode
            ramdisk = 'glance://%s_ramdisk_uuid' % mode
            kernel = 'glance://%s_kernel_uuid' % mode

            config = {
                '%s_ramdisk_by_arch' % mode: {
                    'x86_64': wrong_ramdisk, 'aarch64': ramdisk},
                '%s_kernel_by_arch' % mode: {
                    'x86_64': wrong_kernel, 'aarch64': kernel}
            }
            expected = {
                '%s_ramdisk' % mode: ramdisk,
                '%s_kernel' % mode: kernel
            }

            self.config(group='conductor', **config)

            image_info = redfish_boot._parse_driver_info(task.node)

            for key, value in expected.items():
                self.assertEqual(value, image_info[key])

    def test_parse_driver_info_choose_by_arch_deploy(self):
        self._test_parse_driver_info_choose_by_arch()

    def test_parse_driver_info_choose_by_arch_rescue(self):
        self._test_parse_driver_info_choose_by_arch(mode='rescue')

    def _test_parse_driver_info_choose_by_hierarchy(self, mode='deploy',
                                                    ramdisk_missing=False):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            if mode == 'rescue':
                task.node.provision_state = states.RESCUING

            ramdisk = 'glance://def_%s_ramdisk_uuid' % mode
            kernel = 'glance://def_%s_kernel_uuid' % mode
            ramdisk_by_arch = 'glance://%s_ramdisk_by_arch_uuid' % mode
            kernel_by_arch = 'glance://%s_kernel_by_arch_uuid' % mode

            config = {
                '%s_kernel_by_arch' % mode: {
                    'x86_64': kernel_by_arch},
                '%s_ramdisk' % mode: ramdisk,
                '%s_kernel' % mode: kernel
            }
            if not ramdisk_missing:
                config['%s_ramdisk_by_arch' % mode] = {
                    'x86_64': ramdisk_by_arch}
                expected = {
                    '%s_ramdisk' % mode: ramdisk_by_arch,
                    '%s_kernel' % mode: kernel_by_arch
                }
            else:
                expected = {
                    '%s_ramdisk' % mode: ramdisk,
                    '%s_kernel' % mode: kernel
                }

            self.config(group='conductor', **config)

            image_info = redfish_boot._parse_driver_info(task.node)

            for key, value in expected.items():
                self.assertEqual(value, image_info[key])

    def test_parse_driver_info_choose_by_hierarchy_deploy(self):
        self._test_parse_driver_info_choose_by_hierarchy()

    def test_parse_driver_info_choose_by_hierarchy_rescue(self):
        self._test_parse_driver_info_choose_by_hierarchy(mode='rescue')

    def test_parse_driver_info_choose_by_hierarchy_missing_param_deploy(self):
        self._test_parse_driver_info_choose_by_hierarchy(ramdisk_missing=True)

    def test_parse_driver_info_choose_by_hierarchy_missing_param_rescue(self):
        self._test_parse_driver_info_choose_by_hierarchy(
            mode='rescue', ramdisk_missing=True)

    def test_parse_deploy_info(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.node.instance_info.update(
                {'image_source': 'http://boot/iso',
                 'kernel': 'http://kernel/img',
                 'ramdisk': 'http://ramdisk/img'})

            actual_instance_info = redfish_boot._parse_deploy_info(task.node)

            self.assertEqual(
                'http://boot/iso', actual_instance_info['image_source'])
            self.assertEqual(
                'http://kernel/img', actual_instance_info['kernel'])
            self.assertEqual(
                'http://ramdisk/img', actual_instance_info['ramdisk'])

    def test_parse_deploy_info_exc(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.MissingParameterValue,
                              redfish_boot._parse_deploy_info,
                              task.node)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    def test_validate_local(self, mock_parse_driver_info, mock_get_system):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.instance_info = {}
            mock_get_system.return_value.boot.allowed_values = [
                "UefiHttp", "Hdd"]

            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )
            task.driver.boot.validate(task)

    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    def test_validate_errors_with_lack_of_support(
            self, mock_parse_driver_info, mock_get_system):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.instance_info = {}
            mock_get_system.return_value.boot.allowed_values = ["Hdd"]

            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )
            msg = ("Node %s hardware does not support feature UefiHttp boot, "
                   "which is required based upon the requested configuration."
                   % task.node.uuid)
            self.assertRaisesRegex(
                exception.UnsupportedHardwareFeature,
                msg, task.driver.boot.validate, task)

    @mock.patch.object(redfish_boot.RedfishHttpsBoot, '_validate_hardware',
                       autospec=True)
    @mock.patch.object(deploy_utils, 'get_boot_option', lambda node: 'ramdisk')
    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    @mock.patch.object(deploy_utils, 'validate_image_properties',
                       autospec=True)
    def test_validate_kernel_ramdisk(self, mock_validate_image_properties,
                                     mock_parse_driver_info,
                                     mock_validate_hardware):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.instance_info.update(
                {'kernel': 'kernel',
                 'ramdisk': 'ramdisk',
                 'image_source': 'http://image/source'}
            )

            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.driver.boot.validate(task)

            mock_validate_image_properties.assert_called_once_with(
                task, mock.ANY)
            mock_validate_hardware.assert_called_once_with(mock.ANY, task)

    @mock.patch.object(redfish_boot.RedfishHttpsBoot, '_validate_hardware',
                       autospec=True)
    @mock.patch.object(deploy_utils, 'get_boot_option', lambda node: 'ramdisk')
    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    @mock.patch.object(deploy_utils, 'validate_image_properties',
                       autospec=True)
    def test_validate_boot_iso(self, mock_validate_image_properties,
                               mock_parse_driver_info,
                               mock_validate_hardware):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.instance_info.update(
                {'boot_iso': 'http://localhost/file.iso'}
            )

            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.driver.boot.validate(task)

            mock_validate_image_properties.assert_called_once_with(
                task, mock.ANY)
            mock_validate_hardware.assert_called_once_with(mock.ANY, task)

    @mock.patch.object(redfish_boot.RedfishHttpsBoot, '_validate_hardware',
                       autospec=True)
    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    @mock.patch.object(deploy_utils, 'validate_image_properties',
                       autospec=True)
    def test_validate_correct_vendor(self, mock_validate_image_properties,
                                     mock_parse_driver_info,
                                     mock_validate_hardware):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.instance_info.update(
                {'kernel': 'kernel',
                 'ramdisk': 'ramdisk',
                 'image_source': 'http://image/source'}
            )

            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.node.properties['vendor'] = "Ironic Co."

            task.driver.boot.validate(task)
            mock_validate_hardware.assert_called_once_with(mock.ANY, task)

    @mock.patch.object(redfish_boot.RedfishHttpsBoot, '_validate_hardware',
                       autospec=True)
    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    @mock.patch.object(deploy_utils, 'validate_image_properties',
                       autospec=True)
    def test_validate_missing(self, mock_validate_image_properties,
                              mock_parse_driver_info, mock_validate_hardware):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            self.assertRaises(exception.MissingParameterValue,
                              task.driver.boot.validate, task)
            mock_validate_hardware.assert_called_once_with(mock.ANY, task)

    @mock.patch.object(redfish_utils, 'parse_driver_info', autospec=True)
    def test_validate_inspection(self, mock_parse_driver_info):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.driver_info.update(
                {'deploy_kernel': 'kernel',
                 'deploy_ramdisk': 'ramdisk',
                 'bootloader': 'bootloader'}
            )

            task.driver.boot.validate_inspection(task)

            mock_parse_driver_info.assert_called_once_with(task.node)

    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(image_utils, 'prepare_deploy_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_driver_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_power_action',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_ramdisk_no_debug(
            self, mock_system, mock_boot_mode_utils, mock_node_power_action,
            mock__parse_driver_info,
            mock_prepare_deploy_iso, mock_node_set_boot_device):
        self.config(debug=False)
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING

            mock__parse_driver_info.return_value = {}
            mock_prepare_deploy_iso.return_value = 'image-url'

            task.driver.boot.prepare_ramdisk(task, {})

            mock_node_power_action.assert_called_once_with(
                task, states.POWER_OFF)

            expected_params = {
                'ipa-agent-token': mock.ANY,
                'boot_method': 'vmedia',
            }

            mock_prepare_deploy_iso.assert_called_once_with(
                task, expected_params, 'deploy', {})

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.UEFIHTTP, False)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)

    @mock.patch.object(manager_utils, 'is_fast_track', lambda task: True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(image_utils, 'prepare_deploy_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_driver_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_power_action',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_ramdisk_fast_track(
            self, mock_system, mock_boot_mode_utils, mock_node_power_action,
            mock__parse_driver_info,
            mock_prepare_deploy_iso, mock_node_set_boot_device):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=False) as task:
            task.node.provision_state = states.DEPLOYING

            task.driver.boot.prepare_ramdisk(task, {})

            mock_node_power_action.assert_not_called()
            mock_prepare_deploy_iso.assert_not_called()
            mock_node_set_boot_device.assert_not_called()
            mock_boot_mode_utils.sync_boot_mode.assert_not_called()

    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_floppy_image', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_driver_info', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_clean_up_ramdisk(
            self, mock_system, mock__parse_driver_info,
            mock_cleanup_floppy_image, mock_cleanup_iso_image):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            task.node.driver_info['config_via_removable'] = True

            task.driver.boot.clean_up_ramdisk(task)

            mock_cleanup_iso_image.assert_called_once_with(task)

    @mock.patch.object(redfish_boot.RedfishHttpsBoot,
                       '_clean_up', autospec=True)
    @mock.patch.object(image_utils, 'prepare_boot_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_deploy_info', autospec=True)
    @mock.patch.object(redfish_boot, 'manager_utils', autospec=True)
    @mock.patch.object(redfish_boot, 'deploy_utils', autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_instance_normal_boot(
            self, mock_system, mock_boot_mode_utils, mock_deploy_utils,
            mock_manager_utils, mock__parse_deploy_info,
            mock_prepare_boot_iso, mock_clean_up_instance):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            task.node.driver_internal_info[
                'root_uuid_or_disk_id'] = self.node.uuid

            mock_deploy_utils.get_boot_option.return_value = 'net'

            d_info = {
                'deploy_kernel': 'kernel',
                'deploy_ramdisk': 'ramdisk',
                'bootloader': 'bootloader'
            }

            mock__parse_deploy_info.return_value = d_info

            mock_prepare_boot_iso.return_value = 'image-url'

            task.driver.boot.prepare_instance(task)

            expected_params = {
                'root_uuid': self.node.uuid
            }

            mock_prepare_boot_iso.assert_called_once_with(
                task, d_info, **expected_params)

            mock_manager_utils.node_set_boot_device.assert_called_once_with(
                task, boot_devices.UEFIHTTP, persistent=True)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)
            csb = mock_boot_mode_utils.configure_secure_boot_if_needed
            csb.assert_called_once_with(task)

    @mock.patch.object(redfish_boot.RedfishHttpsBoot,
                       '_clean_up', autospec=True)
    @mock.patch.object(image_utils, 'prepare_boot_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_deploy_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'deploy_utils', autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_instance_ramdisk_boot(
            self, mock_system, mock_boot_mode_utils, mock_deploy_utils,
            mock_node_set_boot_device, mock__parse_deploy_info,
            mock_prepare_boot_iso, mock_clean_up_instance):

        configdrive = 'Y29udGVudA=='
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            task.node.driver_internal_info[
                'root_uuid_or_disk_id'] = self.node.uuid
            task.node.instance_info['configdrive'] = configdrive

            mock_deploy_utils.get_boot_option.return_value = 'ramdisk'

            d_info = {
                'deploy_kernel': 'kernel',
                'deploy_ramdisk': 'ramdisk',
                'bootloader': 'bootloader'
            }
            mock__parse_deploy_info.return_value = d_info

            mock_prepare_boot_iso.return_value = 'image-url'

            task.driver.boot.prepare_instance(task)

            mock_clean_up_instance.assert_called_once_with(mock.ANY, task)

            mock_prepare_boot_iso.assert_called_once_with(task, d_info)

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.UEFIHTTP, persistent=True)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)

    @mock.patch.object(redfish_boot.RedfishHttpsBoot,
                       '_clean_up', autospec=True)
    @mock.patch.object(image_utils, 'prepare_boot_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_deploy_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'deploy_utils', autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_instance_ramdisk_boot_iso(
            self, mock_system, mock_boot_mode_utils, mock_deploy_utils,
            mock_node_set_boot_device, mock__parse_deploy_info,
            mock_prepare_boot_iso,
            mock_clean_up_instance):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            task.node.driver_internal_info[
                'root_uuid_or_disk_id'] = self.node.uuid
            task.node.instance_info['configdrive'] = None

            mock_deploy_utils.get_boot_option.return_value = 'ramdisk'

            d_info = {
                'deploy_kernel': 'kernel',
                'deploy_ramdisk': 'ramdisk',
                'bootloader': 'bootloader'
            }

            mock__parse_deploy_info.return_value = d_info
            mock_prepare_boot_iso.return_value = 'image-url'

            task.driver.boot.prepare_instance(task)

            mock_prepare_boot_iso.assert_called_once_with(task, d_info)

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.UEFIHTTP, persistent=True)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)

    @mock.patch.object(image_utils, 'cleanup_disk_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(image_utils, 'prepare_boot_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_deploy_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'deploy_utils', autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_instance_ramdisk_boot_iso_boot(
            self, mock_system, mock_boot_mode_utils, mock_deploy_utils,
            mock_node_set_boot_device, mock__parse_deploy_info,
            mock_prepare_boot_iso,
            mock_image_cleanup, mock_disk_cleanup):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            i_info = task.node.instance_info
            i_info['boot_iso'] = "super-magic"
            del i_info['configdrive']
            task.node.instance_info = i_info
            mock_deploy_utils.get_boot_option.return_value = 'ramdisk'
            mock__parse_deploy_info.return_value = {}

            mock_prepare_boot_iso.return_value = 'image-url'

            task.driver.boot.prepare_instance(task)

            mock_prepare_boot_iso.assert_called_once_with(task, {})

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.UEFIHTTP, persistent=True)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)
            mock_image_cleanup.assert_called_once_with(task)
            mock_disk_cleanup.assert_not_called()

    @mock.patch.object(image_utils, 'cleanup_disk_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(image_utils, 'prepare_boot_iso', autospec=True)
    @mock.patch.object(redfish_boot, '_parse_deploy_info', autospec=True)
    @mock.patch.object(redfish_boot.manager_utils, 'node_set_boot_device',
                       autospec=True)
    @mock.patch.object(redfish_boot, 'deploy_utils', autospec=True)
    @mock.patch.object(redfish_boot, 'boot_mode_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_prepare_instance_ramdisk_boot_render_configdrive(
            self, mock_system, mock_boot_mode_utils, mock_deploy_utils,
            mock_node_set_boot_device, mock__parse_deploy_info,
            mock_prepare_boot_iso,
            mock_image_cleanup, mock_disk_cleanup):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            task.node.driver_internal_info[
                'root_uuid_or_disk_id'] = self.node.uuid
            task.node.instance_info['configdrive'] = {'meta_data': {}}

            mock_deploy_utils.get_boot_option.return_value = 'ramdisk'

            d_info = {
                'deploy_kernel': 'kernel',
                'deploy_ramdisk': 'ramdisk',
                'bootloader': 'bootloader'
            }
            mock__parse_deploy_info.return_value = d_info

            mock_prepare_boot_iso.return_value = 'image-url'

            task.driver.boot.prepare_instance(task)

            mock_prepare_boot_iso.assert_called_once_with(task, d_info)

            mock_node_set_boot_device.assert_called_once_with(
                task, boot_devices.UEFIHTTP, persistent=True)

            mock_boot_mode_utils.sync_boot_mode.assert_called_once_with(task)
            mock_image_cleanup.assert_called_once_with(task)
            mock_disk_cleanup.assert_not_called()

    @mock.patch.object(boot_mode_utils, 'configure_secure_boot_if_needed',
                       autospec=True)
    @mock.patch.object(boot_mode_utils, 'sync_boot_mode', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(redfish_boot, 'manager_utils', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def _test_prepare_instance_local_boot(
            self, mock_system, mock_manager_utils,
            mock_cleanup_iso_image, mock_sync_boot_mode,
            mock_secure_boot):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            task.node.provision_state = states.DEPLOYING
            task.node.driver_internal_info[
                'root_uuid_or_disk_id'] = self.node.uuid

            task.driver.boot.prepare_instance(task)

            mock_manager_utils.node_set_boot_device.assert_called_once_with(
                task, boot_devices.DISK, persistent=True)
            mock_cleanup_iso_image.assert_called_once_with(task)
            mock_sync_boot_mode.assert_called_once_with(task)
            mock_secure_boot.assert_called_once_with(task)

    def test_prepare_instance_local_whole_disk_image(self):
        self.node.driver_internal_info = {'is_whole_disk_image': True}
        self.node.save()
        self._test_prepare_instance_local_boot()

    def test_prepare_instance_local_boot_option(self):
        instance_info = self.node.instance_info
        instance_info['capabilities'] = '{"boot_option": "local"}'
        self.node.instance_info = instance_info
        self.node.save()
        self._test_prepare_instance_local_boot()

    @mock.patch.object(boot_mode_utils, 'deconfigure_secure_boot_if_needed',
                       autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def _test_clean_up_instance(self, mock_system, mock_cleanup_iso_image,
                                mock_secure_boot):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:

            task.driver.boot.clean_up_instance(task)

            mock_cleanup_iso_image.assert_called_once_with(task)
            mock_secure_boot.assert_called_once_with(task)

    def test_clean_up_instance_only_cdrom(self):
        self._test_clean_up_instance()

    def test_clean_up_instance_cdrom_and_floppy(self):
        driver_info = self.node.driver_info
        driver_info['config_via_removable'] = True
        self.node.driver_info = driver_info
        self.node.save()
        self._test_clean_up_instance()

    @mock.patch.object(boot_mode_utils, 'deconfigure_secure_boot_if_needed',
                       autospec=True)
    @mock.patch.object(deploy_utils, 'get_boot_option', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_disk_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test_clean_up_instance_ramdisk(self, mock_system,
                                       mock_cleanup_iso_image,
                                       mock_cleanup_disk_image,
                                       mock_get_boot_option,
                                       mock_secure_boot):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_get_boot_option.return_value = 'ramdisk'

            task.driver.boot.clean_up_instance(task)

            mock_cleanup_iso_image.assert_called_once_with(task)

            mock_secure_boot.assert_called_once_with(task)
            mock_cleanup_disk_image.assert_not_called()


@mock.patch('oslo_utils.eventletutils.EventletEvent.wait',
            lambda *args, **kwargs: None)
class RedfishVirtualMediaBootViaSystemTestCase(db_base.DbTestCase):

    def setUp(self):
        super(RedfishVirtualMediaBootViaSystemTestCase, self).setUp()
        self.config(enabled_hardware_types=['redfish'],
                    enabled_power_interfaces=['redfish'],
                    enabled_boot_interfaces=['redfish-virtual-media'],
                    enabled_management_interfaces=['redfish'],
                    enabled_inspect_interfaces=['redfish'],
                    enabled_bios_interfaces=['redfish'])
        self.node = obj_utils.create_test_node(
            self.context, driver='redfish', driver_info=INFO_DICT)

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__insert_vmedia_anew(self, mock_system, mock_vmd_sys):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmd_sys.return_value = True
            mock_vmedia_cd = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_CD])
            mock_vmedia_floppy = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])

            mock_manager = mock.MagicMock()

            mock_system.return_value.virtual_media.get_members.return_value = [
                mock_vmedia_cd, mock_vmedia_floppy]

            redfish_boot._insert_vmedia(
                task, [mock_manager], 'img-url', sushy.VIRTUAL_MEDIA_CD)

            mock_vmedia_cd.insert_media.assert_called_once_with(
                'img-url', inserted=True, write_protected=True)

            self.assertFalse(mock_vmedia_floppy.insert_media.call_count)

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__insert_vmedia_anew_dvd(self, mock_system, mock_vmd_sys):
        mock_vmd_sys.return_value = True
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmedia_dvd = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_DVD])

            mock_manager = mock.MagicMock()

            mock_system.return_value.virtual_media.get_members.return_value = [
                mock_vmedia_dvd]

            redfish_boot._insert_vmedia(
                task, [mock_manager], 'img-url', sushy.VIRTUAL_MEDIA_CD)

            mock_vmedia_dvd.insert_media.assert_called_once_with(
                'img-url', inserted=True, write_protected=True)

    @mock.patch('time.sleep', lambda *args, **kwargs: None)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__insert_vmedia_anew_dvd_retry(self, mock_system, mock_vmd_sys):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmd_sys.return_value = True
            mock_vmedia_dvd_1 = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_DVD])

            mock_vmedia_dvd_2 = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_DVD])

            mock_manager = mock.MagicMock()

            def clear_and_raise(*args, **kwargs):
                mock_vmedia_dvd_1.insert_media.side_effect = None
                raise sushy.exceptions.BadRequestError(
                    "POST", 'img-url', mock.MagicMock())
            mock_vmedia_dvd_1.insert_media.side_effect = clear_and_raise
            mock_system.return_value.virtual_media.get_members.return_value = [
                mock_vmedia_dvd_1, mock_vmedia_dvd_2]

            redfish_boot._insert_vmedia(
                task, [mock_manager], 'img-url', sushy.VIRTUAL_MEDIA_CD)

            self.assertEqual(mock_vmedia_dvd_2.insert_media.call_count, 1)

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__insert_vmedia_already_inserted(self, mock_sys, mock_vmd_sys):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmd_sys.return_value = False
            mock_vmedia_cd = mock.MagicMock(
                inserted=True,
                image='img-url',
                media_types=[sushy.VIRTUAL_MEDIA_CD])
            mock_manager = mock.MagicMock()

            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_cd]

            redfish_boot._insert_vmedia(
                task, [mock_manager], 'img-url', sushy.VIRTUAL_MEDIA_CD)

            self.assertFalse(mock_vmedia_cd.insert_media.call_count)

    @mock.patch('time.sleep', lambda *args, **kwargs: None)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__insert_vmedia_while_ejecting(self, mock_system, mock_vmd_sys):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmd_sys.return_value = True
            mock_vmedia_cd = mock.MagicMock(
                inserted=False,
                image='img-url',
                media_types=[sushy.VIRTUAL_MEDIA_CD],
            )
            mock_manager = mock.MagicMock()

            def clear_and_raise(*args, **kwargs):
                mock_vmedia_cd.insert_media.side_effect = None
                raise sushy.exceptions.ServerSideError(
                    "POST", 'img-url', mock.MagicMock())
            mock_vmedia_cd.insert_media.side_effect = clear_and_raise
            mock_system.return_value.virtual_media.get_members.return_value = [
                mock_vmedia_cd]

            redfish_boot._insert_vmedia(
                task, [mock_manager], 'img-url', sushy.VIRTUAL_MEDIA_CD)

            self.assertEqual(mock_vmedia_cd.insert_media.call_count, 2)

    @mock.patch('time.sleep', lambda *args, **kwargs: None)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_utils, 'get_system', autospec=True)
    def test__insert_vmedia_bad_device(self, mock_sys, mock_vmd_sys):
        mock_vmd_sys.return_value = True
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmedia_floppy = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])
            mock_system = mock.MagicMock()
            mock_manager = mock.MagicMock()

            mock_system.virtual_media.get_members.return_value = [
                mock_vmedia_floppy]

            self.assertRaises(
                exception.InvalidParameterValue,
                redfish_boot._insert_vmedia,
                task, [mock_manager], 'img-url', sushy.VIRTUAL_MEDIA_CD)

    @mock.patch.object(image_utils, 'cleanup_disk_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_boot, 'redfish_utils', autospec=True)
    def test_eject_vmedia_everything(self, mock_redfish_utils, mock_vmd_sys,
                                     mock_cleanup_iso, mock_cleanup_disk):
        mock_vmd_sys.return_value = True
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmedia_cd = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_CD])
            mock_vmedia_floppy = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])
            mock_vmedia_dvd = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_DVD])

            mock_manager = mock.MagicMock()
            mock_system = mock.MagicMock()

            mock_redfish_utils.get_system.return_value = mock_system
            mock_system.managers = [
                mock_manager]

            mock_system.virtual_media.get_members.return_value = [
                mock_vmedia_cd, mock_vmedia_floppy, mock_vmedia_dvd]

            redfish_boot.eject_vmedia(task)

            mock_vmedia_cd.eject_media.assert_called_once_with()
            mock_vmedia_floppy.eject_media.assert_called_once_with()
            mock_vmedia_dvd.eject_media.assert_called_once_with()
            mock_cleanup_iso.assert_called_once_with(task)
            mock_cleanup_disk.assert_called_once_with(task,
                                                      prefix='configdrive')

    @mock.patch.object(image_utils, 'cleanup_disk_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_manager', autospec=True)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_boot, 'redfish_utils', autospec=True)
    def test_eject_vmedia_from_all_resources(self, mock_redfish_utils,
                                             mock_vmd_sys, mock_vmd_mg,
                                             mock_cleanup_iso,
                                             mock_cleanup_disk):
        mock_vmd_sys.return_value = True
        mock_vmd_mg.return_value = True
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmedia_cd = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_CD])
            mock_vmedia_floppy = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])
            mock_vmedia_dvd = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_DVD])

            mock_manager = mock.MagicMock()
            mock_system = mock.MagicMock()

            mock_redfish_utils.get_system.return_value = mock_system
            mock_system.managers = [
                mock_manager]

            mock_system.virtual_media.get_members.return_value = [
                mock_vmedia_floppy, mock_vmedia_dvd]
            mock_manager.virtual_media.get_members.return_value = [
                mock_vmedia_cd]

            redfish_boot.eject_vmedia(task)

            mock_vmedia_cd.eject_media.assert_called_once_with()
            mock_vmedia_floppy.eject_media.assert_called_once_with()
            mock_vmedia_dvd.eject_media.assert_called_once_with()
            mock_cleanup_iso.assert_called_once_with(task)
            mock_cleanup_disk.assert_called_once_with(task,
                                                      prefix='configdrive')

    @mock.patch.object(image_utils, 'cleanup_disk_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_boot, 'redfish_utils', autospec=True)
    def test_eject_vmedia_specific(self, mock_redfish_utils, mock_vmd_sys,
                                   mock_cleanup_iso, mock_cleanup_disk):
        mock_vmd_sys.return_value = True
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmedia_cd = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_CD])
            mock_vmedia_floppy = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])

            mock_system = mock.MagicMock()
            mock_manager = mock.MagicMock()

            mock_redfish_utils.get_system.return_value = mock_system
            mock_system.managers = [
                mock_manager]
            mock_system.virtual_media.get_members.return_value = [
                mock_vmedia_cd, mock_vmedia_floppy]

            redfish_boot.eject_vmedia(task, sushy.VIRTUAL_MEDIA_CD)

            mock_vmedia_cd.eject_media.assert_called_once_with()
            self.assertFalse(mock_vmedia_floppy.eject_media.call_count)
            mock_cleanup_iso.assert_called_once_with(task)
            mock_cleanup_disk.assert_not_called()

    @mock.patch.object(image_utils, 'cleanup_disk_image', autospec=True)
    @mock.patch.object(image_utils, 'cleanup_iso_image', autospec=True)
    @mock.patch.object(redfish_boot, 'redfish_utils', autospec=True)
    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_boot.LOG, 'debug', autospec=True)
    @mock.patch.object(redfish_boot.LOG, 'info', autospec=True)
    def test_eject_vmedia_with_dvd_cisco_ucs(self, mock_log_info,
                                             mock_log_debug,
                                             mock_vmd_sys,
                                             mock_redfish_utils,
                                             mock_cleanup_iso,
                                             mock_cleanup_disk):
        mock_vmd_sys.return_value = True
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmedia_dvd_1 = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_DVD])
            mock_vmedia_dvd_2 = mock.MagicMock(
                inserted=True,
                media_types=[sushy.VIRTUAL_MEDIA_DVD])

            mock_system = mock.MagicMock()
            mock_manager = mock.MagicMock()

            mock_redfish_utils.get_system.return_value = mock_system
            mock_system.managers = [
                mock_manager]
            mock_system.virtual_media.get_members.return_value = [
                mock_vmedia_dvd_1, mock_vmedia_dvd_2]

            redfish_boot.eject_vmedia(task, sushy.VIRTUAL_MEDIA_CD)

            mock_vmedia_dvd_1.eject_media.assert_called_once_with()
            mock_vmedia_dvd_2.eject_media.assert_called_once_with()

            self.assertEqual(mock_log_info.call_count, 2)
            self.assertEqual(mock_log_debug.call_count, 3)
            mock_cleanup_iso.assert_called_once_with(task)
            mock_cleanup_disk.assert_not_called()

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_boot, 'redfish_utils', autospec=True)
    def test_eject_vmedia_not_inserted(self, mock_redfish_utils, mock_vmd_sys):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmd_sys.return_value = True
            mock_vmedia_cd = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_CD])
            mock_vmedia_floppy = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])

            mock_manager = mock.MagicMock()
            mock_system = mock.MagicMock()

            mock_system.virtual_media.get_members.return_value = [
                mock_vmedia_cd, mock_vmedia_floppy]

            mock_redfish_utils.get_system.return_value.managers = [
                mock_manager]

            redfish_boot.eject_vmedia(task)

            self.assertFalse(mock_vmedia_cd.eject_media.call_count)
            self.assertFalse(mock_vmedia_floppy.eject_media.call_count)

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    @mock.patch.object(redfish_boot, 'redfish_utils', autospec=True)
    def test_eject_vmedia_unknown(self, mock_redfish_utils, mock_vmd_sys):

        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            mock_vmd_sys.return_value = True
            mock_vmedia_cd = mock.MagicMock(
                inserted=False,
                media_types=[sushy.VIRTUAL_MEDIA_CD])

            mock_manager = mock.MagicMock()
            mock_system = mock.MagicMock()

            mock_system.virtual_media.get_members.return_value = [
                mock_vmedia_cd]

            mock_redfish_utils.get_system.return_value.managers = [
                mock_manager]

            redfish_boot.eject_vmedia(task)

            self.assertFalse(mock_vmedia_cd.eject_media.call_count)

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    def test__has_vmedia_device(self, mock_vmd_sys):
        mock_vmd_sys.return_value = True
        mock_vmedia_cd = mock.MagicMock(
            inserted=False,
            media_types=[sushy.VIRTUAL_MEDIA_CD])
        mock_vmedia_floppy = mock.MagicMock(
            inserted=False,
            media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])

        mock_manager = mock.MagicMock()

        mock_system = mock.MagicMock()
        mock_system.virtual_media.get_members.return_value = [
            mock_vmedia_cd, mock_vmedia_floppy]

        self.assertEqual(
            sushy.VIRTUAL_MEDIA_CD,
            redfish_boot._has_vmedia_device(
                [mock_manager], sushy.VIRTUAL_MEDIA_CD,
                system=mock_system))

        self.assertFalse(
            redfish_boot._has_vmedia_device(
                [mock_manager], sushy.VIRTUAL_MEDIA_CD,
                inserted=True, system=mock_system))

        self.assertFalse(
            redfish_boot._has_vmedia_device(
                [mock_manager], sushy.VIRTUAL_MEDIA_USBSTICK,
                system=mock_system))

    @mock.patch.object(redfish_boot, '_has_vmedia_via_systems', autospec=True)
    def test__has_vmedia_device_inserted(self, mock_vmd_sys):
        mock_vmd_sys.return_value = True
        mock_vmedia_cd = mock.MagicMock(
            inserted=False,
            media_types=[sushy.VIRTUAL_MEDIA_CD])
        mock_vmedia_floppy = mock.MagicMock(
            inserted=True,
            media_types=[sushy.VIRTUAL_MEDIA_FLOPPY])

        mock_manager = mock.MagicMock()
        mock_system = mock.MagicMock()
        mock_system.virtual_media.get_members.return_value = [
            mock_vmedia_cd, mock_vmedia_floppy]

        self.assertEqual(
            sushy.VIRTUAL_MEDIA_FLOPPY,
            redfish_boot._has_vmedia_device(
                [mock_manager], sushy.VIRTUAL_MEDIA_FLOPPY,
                inserted=True, system=mock_system))



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from ironic.common.i18n import _

opts = [
    cfg.StrOpt('storage',
               default='swift',
               help=_('Configuration mold storage location. Supports "swift" '
                      'and "http". By default "swift".')),
    cfg.StrOpt('user',
               help=_('User for "http" Basic auth. By default set empty.')),
    cfg.StrOpt('password',
               secret=True,
               help=_('Password for "http" Basic auth. By default set '
                      'empty.')),
    cfg.IntOpt('retry_attempts',
               default=3,
               help=_('Retry attempts for saving or getting configuration '
                      'molds.')),
    cfg.IntOpt('retry_interval',
               default=3,
               help=_('Retry interval for saving or getting configuration '
                      'molds.'))
]


def register_opts(conf):
    conf.register_opts(opts, group='molds')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gather and distinguish PCI devices from plugin_data."""

import collections
import json

from oslo_config import cfg
from oslo_log import log as logging

from ironic.common import utils
from ironic.drivers.modules.inspector.hooks import base


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def _parse_pci_alias_entries():
    parsed_pci_devices = []

    for pci_alias_entry in CONF.inspector.pci_device_alias:
        try:
            parsed_entry = json.loads(pci_alias_entry)
            if set(parsed_entry) != {'vendor_id', 'product_id', 'name'}:
                raise KeyError('The "pci_device_alias" entry should contain '
                               'exactly the "vendor_id", "product_id" and '
                               '"name" keys.')
            parsed_pci_devices.append(parsed_entry)
        except (ValueError, KeyError) as exc:
            LOG.error("Error parsing 'pci_device_alias' option: %s", exc)

    return {(dev['vendor_id'], dev['product_id']): dev['name']
            for dev in parsed_pci_devices}


class PciDevicesHook(base.InspectionHook):
    """Hook to count various PCI devices, and set the node's capabilities.

    This information can later be used by nova for node scheduling.
    """
    _aliases = _parse_pci_alias_entries()

    def _found_pci_devices_count(self, found_pci_devices):
        return collections.Counter([(dev['vendor_id'], dev['product_id'])
                                    for dev in found_pci_devices
                                    if (dev['vendor_id'], dev['product_id'])
                                    in self._aliases])

    def __call__(self, task, inventory, plugin_data):
        """Update node capabilities with PCI devices."""

        if 'pci_devices' not in plugin_data:
            if CONF.inspector.pci_device_alias:
                LOG.warning('No information about PCI devices was received '
                            'from the ramdisk.')
            return

        alias_count = {self._aliases[id_pair]: count for id_pair, count in
                       self._found_pci_devices_count(
                           plugin_data['pci_devices']).items()}
        if alias_count:
            LOG.info('Found the following PCI devices: %s', alias_count)

            old_capabilities = task.node.properties.get('capabilities')
            LOG.debug('Old capabilities for node %s: %s', task.node.uuid,
                      old_capabilities)
            new_capabilities = utils.get_updated_capabilities(old_capabilities,
                                                              alias_count)
            task.node.set_property('capabilities', new_capabilities)
            LOG.debug('New capabilities for node %s: %s', task.node.uuid,
                      new_capabilities)
            task.node.save()



#
# Copyright 2014 OpenStack Foundation
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from urllib import parse as urlparse

import openstack
from openstack.connection import exceptions as openstack_exc
from oslo_log import log

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import keystone
from ironic.conf import CONF

LOG = log.getLogger(__name__)

_SWIFT_SESSION = None


def get_swift_session():
    global _SWIFT_SESSION
    if not _SWIFT_SESSION:
        auth = keystone.get_auth('swift')
        _SWIFT_SESSION = keystone.get_session('swift', auth=auth)
    return _SWIFT_SESSION


class SwiftAPI(object):
    """API for communicating with Swift."""

    def __init__(self):
        """Initialize the connection with swift"""
        self.connection = openstack.connection.Connection(
            session=get_swift_session(),
            oslo_conf=CONF)

    def create_object(self, container, obj, filename,
                      object_headers=None):
        """Uploads a given file to Swift.

        :param container: The name of the container for the object.
        :param obj: The name of the object in Swift
        :param filename: The file to upload, as the object data
        :param object_headers: the headers for the object to pass to Swift
        :returns: The Swift UUID of the object
        :raises: SwiftOperationError, if any operation with Swift fails.
        """
        try:
            self.connection.create_container(container)
        except openstack_exc.OpenStackCloudException as e:
            operation = _("put container")
            raise exception.SwiftOperationError(operation=operation, error=e)

        if object_headers is None:
            object_headers = {}

        try:
            obj_uuid = self.connection.create_object(
                container, obj, filename=filename, **object_headers)
        except openstack_exc.OpenStackCloudException as e:
            operation = _("put object")
            raise exception.SwiftOperationError(operation=operation,
                                                error=e)

        return obj_uuid

    def create_object_from_data(self, obj, data, container):
        """Uploads a given string to Swift.

        :param obj: The name of the object in Swift
        :param data: string data to put in the object
        :param container: The name of the container for the object.
            Defaults to the value set in the configuration options.
        :returns: The Swift UUID of the object
        :raises: utils.Error, if any operation with Swift fails.
        """
        try:
            self.connection.create_container(container)
        except openstack_exc.OpenStackCloudException as e:
            operation = _("put container")
            raise exception.SwiftOperationError(operation=operation, error=e)

        try:
            obj_uuid = self.connection.create_object(
                container, obj, data=data)
        except openstack_exc.OpenStackCloudException as e:
            operation = _("put object")
            raise exception.SwiftOperationError(operation=operation, error=e)

        return obj_uuid

    def get_temp_url_key(self):
        """Get the best temporary url key from the account headers."""
        return self.connection.object_store.get_temp_url_key()

    def get_temp_url(self, container, obj, timeout):
        """Returns the temp url for the given Swift object.

        :param container: The name of the container in which Swift object
            is placed.
        :param obj: The name of the Swift object.
        :param timeout: The timeout in seconds after which the generated url
            should expire.
        :returns: The temp url for the object.
        :raises: SwiftOperationError, if any operation with Swift fails.
        """
        endpoint = keystone.get_endpoint('swift', session=get_swift_session())
        parse_result = urlparse.urlparse(endpoint)
        swift_object_path = '/'.join((parse_result.path, container, obj))
        temp_url_key = self.get_temp_url_key()
        if not temp_url_key:
            raise exception.MissingParameterValue(_(
                'Swift temporary URLs require a shared secret to be '
                'created. You must provide pre-generate the key on '
                'the project used to access Swift.'))
        url_path = self.generate_temp_url(
            swift_object_path, timeout, 'GET', temp_url_key=temp_url_key)
        return urlparse.urlunparse(
            (parse_result.scheme, parse_result.netloc, url_path,
             None, None, None))

    def generate_temp_url(self, path, timeout, method, temp_url_key):
        """Returns the temp url for a given path"""
        return self.connection.object_store.generate_temp_url(
            path, timeout, method, temp_url_key=temp_url_key)

    def get_object(self, object, container):
        """Downloads a given object from Swift.

        :param object: The name of the object in Swift
        :param container: The name of the container for the object.
            Defaults to the value set in the configuration options.
        :returns: Swift object
        :raises: utils.Error, if the Swift operation fails.
        """
        try:
            headers, obj = self.connection.get_object(
                object, container=container)
        except openstack_exc.OpenStackCloudException as e:
            operation = _("get object")
            raise exception.SwiftOperationError(operation=operation, error=e)

        return obj

    def delete_object(self, container, obj):
        """Deletes the given Swift object.

        :param container: The name of the container in which Swift object
            is placed.
        :param obj: The name of the object in Swift to be deleted.
        :raises: SwiftObjectNotFoundError, if object is not found in Swift.
        :raises: SwiftOperationError, if operation with Swift fails.
        """
        try:
            self.connection.delete_object(container, obj)
        except openstack_exc.OpenStackCloudException as e:
            operation = _("delete object")
            if isinstance(e, openstack_exc.ResourceNotFound):
                raise exception.SwiftObjectNotFoundError(obj=obj,
                                                         container=container,
                                                         operation=operation)

            raise exception.SwiftOperationError(operation=operation, error=e)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for the API /shards/ methods.
"""

from http import client as http_client
import uuid

from ironic.api.controllers import base as api_base
from ironic.api.controllers import v1 as api_v1
from ironic.tests.unit.api import base as test_api_base
from ironic.tests.unit.objects import utils as obj_utils


class TestListShards(test_api_base.BaseApiTest):
    headers = {api_base.Version.string: str(api_v1.max_version())}

    def _create_test_shard(self, name, count):
        for i in range(count):
            obj_utils.create_test_node(
                self.context, uuid=uuid.uuid4(), shard=name)

    def test_empty(self):
        data = self.get_json('/shards', headers=self.headers)
        self.assertEqual([], data['shards'])

    def test_one_shard(self):
        shard = 'shard1'
        count = 1
        self._create_test_shard(shard, count)
        data = self.get_json('/shards', headers=self.headers)
        self.assertEqual(shard, data['shards'][0]['name'])
        self.assertEqual(count, data['shards'][0]['count'])

    def test_multiple_shards(self):
        for i in range(0, 6):
            self._create_test_shard('shard{}'.format(i), i)
        data = self.get_json('/shards', headers=self.headers)
        self.assertEqual(5, len(data['shards']))

    def test_nodes_but_no_shards(self):
        self._create_test_shard(None, 5)
        data = self.get_json('/shards', headers=self.headers)
        self.assertEqual("None", data['shards'][0]['name'])
        self.assertEqual(5, data['shards'][0]['count'])

    def test_fail_wrong_version(self):
        headers = {api_base.Version.string: '1.80'}
        self._create_test_shard('shard1', 1)
        result = self.get_json(
            '/shards', expect_errors=True, headers=headers)
        self.assertEqual(http_client.NOT_FOUND, result.status_int)

    def test_fail_get_one(self):
        # We do not implement a get /v1/shards/<shard> endpoint
        # validate it errors properly
        self._create_test_shard('shard1', 1)
        result = self.get_json(
            '/shards/shard1', expect_errors=True, headers=self.headers)
        self.assertEqual(http_client.NOT_FOUND, result.status_int)

    def test_fail_get_one_wrong_version(self):
        self._create_test_shard('shard1', 1)
        # We should get the same response if we request with the wrong version
        headers = {api_base.Version.string: '1.80'}
        result = self.get_json(
            '/shards/shard1', expect_errors=True, headers=headers)
        self.assertEqual(http_client.NOT_FOUND, result.status_int)

    def test_fail_post(self):
        result = self.post_json(
            '/shards', {}, expect_errors=True, headers=self.headers)
        self.assertEqual(http_client.METHOD_NOT_ALLOWED, result.status_int)

    def test_fail_put(self):
        result = self.put_json(
            '/shards', {}, expect_errors=True, headers=self.headers)
        self.assertEqual(http_client.METHOD_NOT_ALLOWED, result.status_int)
